*************************************************************
* (C) COPYRIGHT TEXAS INSTRUMENTS, INC. 1996                *
*************************************************************
* MODULE NAME: AICDSK50.ASM                                 *
*                                                           *
* FUNCTION:    Sampling an analog input signal              *
*              and sending it to analog output              *                                
*              with an overall gain of 1                    * 
*                                                           *
* DESCRIPTION: This program initializes the TMS320C50 and   *
*              TLC32040 on the C5x DSK Board.               *
*                                                           *
*              BAND-PASS: ON                                *
*              Low cut-off frequency: 200 Hz                *
*              High cut-off frequency: 2.5 kHz              *
*              Sampling frequency: 8 kHz                    *
*                                                           *
* OPERATING CONDITIONS:                                     *
*                                                           *
*         *****************************************         *
*         * CAUTION: MAXIMUM INPUT VOLTAGE: 3 V   *         *
*         *****************************************         *         
*                                                           *
*               ANALOG INPUT <- Function Generator          *
*                                                           *
*               ANALOG OUTPUT -> Oscilloscope               *
*                                                           *
*                                                           *
* ASSEMBLING AND DIRECT EXECUTION:                          *
*                                                           *
* You must assemble and link this program with the          * 
* Assembler/linker provided while purchasing the DSK.       *
*                                                           *
*               DSK5A AICDSK50.ASM ->generates AICDSK50.DSK *
*               DSK5L AICDSK50.DSK                          *
*                                                           *
*************************************************************
*                                                           *
* ASSEMBLING AND EXECUTION THROUGH DEBUGGER:                *
*                                                           *
* You must assemble and link this program with the          * 
* Assembler/linker provided while purchasing the DSK        *
*                                                           *
*               DSK5A AICDSK50.ASM ->generates AICDSK50.DSK *
*               DSK5D                                       *
*                                                           *
* Then, reset the DSK board by typing R                     *
* Then, load the DSK program by typing LD then AIC          *
* Then, execute the DSK program by typing XR                *
*************************************************************
	 
	 .mmregs                ; declaration of memory mapped registers


****************************************************************************
* In this section, TA, RA, TB, RB, AIC_CTR registers are declared in data  *
* memory and initialized.                                                  *
****************************************************************************
	.ds     0f00h           ; User's space data memory
TA       .word    25             ; Fcut = 2.5 KHz
RA       .word    25             ; 
TAp      .word    31            ;
RAp      .word    31            ;
TB       .word    25            ; Fs = 8 kHz
RB       .word    25            ; 
AIC_CTR  .word    19h           ; Overall gain=1, sync, loopback disabled 
				; band-pass filter enabled

;---------------------------------------------------------------;
; Bit definition of the AIC_CTR control register                ;
;---------------------------------------------------------------;
;       +----------+-----------------+
;       |  G1  G0  |  SY  AX  LB  BP |                    G1  G0   Gain
;       +----------+-----------------+                   ---------------
;          |   |      |   |   |   +-- Band-Pass Filter    0   0    1/2
;          |   |      |   |   +------ Loopback            1   1    1/2
;          |   |      |   +---------- Aux                 0   1    1
;  Gain ---+---+      +-------------- Synch               1   0    2


*************************************************************************
* Set up the interrupt vector table in program memory and make the link *
* between the interrupt vector and its associated interrupt subroutine. *
* The interrupt vector table starts at 0800h, but location of serial    *
* port receive interrupt RINT is at 080Ah since it is the 5th in the    *
* interrupt table.                                                      *
*************************************************************************
	 .ps     080ah          ; Interrupt Vectors memory
rint:    B       RECEIVE        ; 080Ah: Serial port receive interrupt
xint:    B       TRANSMIT       ; 080Ch: Serial port transmit interrupt




*************************
* START OF MAIN PROGRAM *
*************************


	 .ps 0a00h              ; User's Program memory
	 .entry                 ; entry point of the program
********************************
* TMS320C50 DSP Initialization *
********************************
START:   SETC    INTM           ; Disable interrupts
	 LDP     #0             ; Set data page pointer
	 OPL     #0834h,PMST    ; Interrupt table starts at 0800h
				; Program and data ON-CHIP SARAM 
				; (RAM=1 and OVLY=1) 
				; microcomputer mode MP/MC=0
				; NDX=1, TRM=0 => C5X not enhanced mode
	 LACC    #0
	 SAMM    CWSR           ; zero wait-state access to memory
	 SAMM    PDWSR          ;

* Reset AIC by writing to PA2 (address >52) to DSK
	 SPLK    #022h,IMR      ; using XINT for synch (and INT2 for debugger)
	 CALL    AICINIT        ; call to subroutine of initialization
				; of the AIC

*****************************************************************************
* After the next 4 instructions, the serial port receive interrupt is       *
* enabled. When RINT is triggered, the receive interrupt subroutine         *
* (RECEIVE: ) is executed: it stores the data contained in the DRR register *
* to Accumulator  then stores it to the DXR register.                       *
*****************************************************************************
	 CLRC    OVM            ; OVM = 0 - Overflow mode
	 SPM     0              ; PM = 0 no shift after product reg.
	 SPLK    #012h,IMR      ; RINT only (and INT2 for debugger)
	 CLRC   INTM            ; enable interrupt

*************************************************
* Infinite loop - Waiting for reception of data *
*                                               *
* NB: the RINT interrupt is enabled             *
*************************************************
WAIT    IDLE                    ; wait until interrupt
	B      WAIT             ; wait for input data


***********************        
* END OF MAIN PROGRAM *
***********************



**********************
* INTERRUPT ROUTINES * 
**********************

**************************************
* Receive Interrupt Service Routine  *
**************************************
RECEIVE:
	 LAMM    DRR            ; read data from DRR
	 AND    #0FFFCh         ; 14-bit resolution of the AIC
	 SAMM    DXR            ; send data to DXR (primary communic.) 
	 RETE

**************************************
* Transmit Interrupt Service Routine *
**************************************
TRANSMIT:
	RETE


*************
* FUNCTIONS *
*************

**************************************************************************
* The AICINIT routine initializes the Serial Port and the TLC320C40 AIC, *
* by the following steps:                                                *
* Step 1: Set the Timer frequency (10.000 MHz for AIC MCLK).             *
* Step 2: Configure frame synchronization signals.                       *
* Step 3: Initialisation of the Serial port                              *
* Step 4: Initialisation of the AIC                                      *
*                                                                        *
* NB: AIC Reset line is connected to the /BR (Bus Request) pin of the    * 
*     'C50. This pin is driven low when external global memory is        *
*     accessed. Therefore to reset the AIC, global memory must be        * 
*     defined and accessed to drive the /BR pin low for at least 800ns.  *
**************************************************************************
AICINIT: SPLK    #20h,TCR       ; To generate 10.000 MHz from Tout
	 SPLK    #01h,PRD       ; for AIC master clock
	 MAR     *,AR0
	 LACC    #0008h         ; Non continuous mode, FSM=1, FSX as input 
	 SACL    SPC            ; reset the serial port
	 LACC    #00c8h         ; 16 bit words
	 SACL    SPC            ; bring the serial port out of reset
	 LACC    #080h          ; init 8000h-FFFFh as global memory
	 SACH    DXR            ; Dummy word sent to clear SPC register
	 SACL    GREG           ; Store to Global Memory Alloc Reg.
	 LAR     AR0,#0FFFFh    ; Use AR0 to point to location FFFFh
	 RPT     #10000         ; Access global memory 10000 times
	 LACC    *,0,AR0        ; to drive pin low for duration
	 SACH    GREG           ; Restore GREG to 0000
	 
*****************************************************************************
* The 3 blocks of instructions below send the parameters TA, TB and AIC_CTR *
* to the AIC for a 16 kHz sample rate with a gain setting of 1, BP enabled. *                                            
* The contents of TA and TB ands AIC_CTR registers are each shifted 9 bits  *
* left then added to 2 control bits to configure a complete 16 bit AIC DX   *
* DATA word before its transmission to the AIC.                             *
*                                                                           *
* NB: Sending such a parameter requires a "secondary communication", hence  *
*     the call to the AIC_2ND subroutine.                                   * 
*****************************************************************************         
	 ;-----------------------
	 LDP     #TA            ; Initialization of the data-page pointer
	 SETC    SXM            ; Set sign extension mode
	 LACC    TA,9           ; TA is shifted left 9 bits
	 ADD     RA,2           ; RA is shifted left 2 bits
	 CALL    AIC_2ND        ; call to AIC_2ND subroutine
	 ;-----------------------
	 LDP     #TB            ; Initialization of the data-page pointer
	 LACC    TB,9           ; TB is shifted left 9 bits
	 ADD     RB,2           ; RB is shifted left 2 bits
	 ADD     #02h           ;
	 CALL    AIC_2ND        ;
	 ;-----------------------
	 LDP     #AIC_CTR       ; Initialization of the data-page pointer
	 LACC    AIC_CTR,2      ; AIC_CTR is shifted left 2 bits.
	 ADD     #03h           ;
	 CALL    AIC_2ND        ;
	 RET                    ;

*********************************************************************
* The AIC_2ND routine initializes a secondary transmission, by the  * 
* following steps:                                                  *
* Step 1: enable serial port transmit interrupt                     *
* Step 2: send 03h to initiate a secondary transmission             *
* Step 3: wait until ready for new data (idle)                      * 
* Step 4: send AIC parameter                                        *
* Step 5: disable serial port transmit interrupt                    *
*                                                                   *
* NB:  The lower part of the accumulator contains the contents of   *
*      the AIC DX DATA configuration word. The higher part contains *          
*      a control word which initiates secondary transmission mode   *
*********************************************************************
AIC_2ND:
	 LDP    #0
	 SACH    DXR            ; dummy word sent to clear SPC register
	 CLRC    INTM           ; enable interrupt
	 IDLE                   ; wait for transmit interrrupt
	 ADD     #6h,15         ; 0000 0000 0000 0011 | AIC DX DATA WORD
	 SACH    DXR            ; Send 0003h to initiate 2nd transmission
	 IDLE
	 SACL    DXR            ; Send AIC DX data word
	 IDLE
	 LACL    #0
	 SACL    DXR            ; make sure the word got sent
	 IDLE
	 SETC    INTM           ; disable interrupt
	 RET
	.end
