*******************************************************************************
*       (C) COPYRIGHT CURTIN UNIVERSITY, PERTH, WESTERN AUSTRALIA.            *   
*       Written by Ivar Siewert 930609A (February to June 1998).              *
*******************************************************************************
*                                                                             *
*  PROGRAM NAME: AUSTRALIAN CURRENCY NOTE READER FOR THE VISION IMPAIRED      *
*                                                                             *
*  DESCRIPTION: This program for the Texas Instruments TMS320C50 Digital      *
*               Signal Processor (DSP) is designed to recognise the           *
*               image of the clear window on Australian currency notes.       *
*               The image is produced by a Texas Instruments TSL218           *
*               linear array with a length of 512 pixels.                     *
*                                                                             *
*******************************************************************************

            .mmregs                 ;define global symbolic names for
                                    ; TMS320C5x registers.

*******************************************************************************
*                             GLOBAL CONSTANTS                                *
*******************************************************************************

DATA_PORT   .set  50h               ;memory-mapped data Port Address 0.
FREQUENCY   .set  1820              ;period of one cycle of TONE.
DURATION    .set  100               ;number of cycles of TONE.
PAUSE       .set  100               ;length of pause between BEEPs in ms.

*******************************************************************************
*                             GLOBAL VARIABLES                                *
*******************************************************************************

            .ds   2B80h
FLAGS       .word 0                 ;Explanation of flags:
                                    ;bit 0      SI flag

No_OF_BEEPS .word 3                 ;number of BEEPS.

*******************************************************************************
*                             INTERRUPT VECTORS                               *
*******************************************************************************

            ;set interrupt vectors

            .ps   0803h
            .word 0980h             ;store INT1 ISR address (CLKL signal).
            .ps   0802h
            .word 7980h             ;store branch instruction.
            .ps   0807h
            .word 098Fh             ;store INT3 ISR address (SO signal).
            .ps   0806h
            .word 7980h             ;store branch instruction.
      
*******************************************************************************
*                         INTERRUPT SERVICE ROUTINES                          *
*******************************************************************************

            .ps   0980h             ;assemble code starting at this address.

            ;CLocK Low Interrupt Service Routine

            ;This routine is triggered by external interrupt INT1.
            ;The SI flag is first tested to see if this is the beginning of
            ; a new scan. If so, the SI signal is set high, the flag is
            ; cleared and the routine is exited. A zero SI flag means that
            ; valid pixel data needs to read and stored.
            ;The Analogue Out voltage is valid for a period of half a clock
            ; cycle immediately following a falling clock edge. This is
            ; when the analogue to digital conversion must take place.

CLKL_ISR    LDP   #87
            BIT   FLAGS,15          ;test SI flag (bit 0).
            BCND  DO_AtoD,NTC       ;continue if flag set, otherwise perform
                                    ; A to D conversion and store.
            LDP   #0
            OPL   #8000h,DATA_PORT  ;set DAT15 high (SI signal).
            LDP   #87
            APL   #0FFFEh,FLAGS     ;clear SI flag.
            RETE

DO_AtoD     LDP   #0
            APL   #7FFFh,DATA_PORT  ;clear DAT15 (SI signal).

            ;test SAMPLE flag
            ;perform A to D conversion and store.
            ;if no time, can store data at beginning of next cycle.
            RETE
                                    ;Don't forget to update interrupt vectors.

;------------------------------------------------------------------------------

            ;Serial Out SIGnal Interrupt Service Routine.
            ;This routine is triggered by external interrupt INT3.

SO_SIG_ISR  LDP   #0
            RPT   #20               ;wait ? us.
            OPL   #4000h,DATA_PORT  ;set DAT14, turn LEDs on.
            APL   #0BFFFh,DATA_PORT ;clear DAT14, turn LEDs off.
            LDP   #87
            OPL   #01h,FLAGS        ;set SI flag for next scan.
            RETE
                                    ;Don't forget to update interrupt vectors.

******************************************************************************
*                             SUBROUTINES                                    *
******************************************************************************

            ;TMS320C50 Initialisation.

            ;PMST = Processor Mode STatus register
            ;PRD = timer PeRioD register
            ;TCR = Timer Control Register
            ;IFR = Interrupt Flag Register
            ;IMR = Interupt Mask Register

INITIALISE  LDP   #0                ;data pointer points to page 0.
            SPLK  #10h,TCR          ;timer stopped with no prescale.
            SPLK  #19,PRD           ;load timer for 1 us period (20MHz/[19+1]).
            OPL   #05h,IMR          ;unmask INT1 and INT3 external interrupts.
            CLRC  INTM              ;enable maskable interrupts.
            RET

;------------------------------------------------------------------------------

            ;Produce a single beep

ONE_BEEP    LAR   AR0,#DURATION     ;load AR0 with number of cycles.
            LDP   #0
REPEAT      BANZ  ONE_CYCLE,*-,AR0
            RET

ONE_CYCLE   RPT   #FREQUENCY
            OPL   #2000h,DATA_PORT  ;take DAT13 high for half the cycle.
            RPT   #FREQUENCY
            APL   #0DFFFh,DATA_PORT ;take DAT13 low again for other half.
            B     REPEAT

;------------------------------------------------------------------------------

            ;1 ms delay.

DELAY       RPT   #20000
            NOP
            RET

;------------------------------------------------------------------------------

            ;explanation

BEEPS       ;LDP   #87
            LAR   AR1,#0005h

BACK        BANZ  BEEP&PAUSE,*-,AR1
            RET

BEEP&PAUSE  CALL  ONE_BEEP
            LAR   AR2,#PAUSE

DELAY_AGAIN CALL  DELAY
            BANZ  DELAY_AGAIN,*-,AR2
            B     BACK

;------------------------------------------------------------------------------

            ;explanation here

START_CCD   LDP   #87
            OPL   #01h,FLAGS        ;set SI flag (bit 0) to start scan.
            LDP   #0
            SPLK  #20h,TCR          ;reload and start timer.
                                    ;at least 12 cycles to purge pixels.
                                    ; then auto-calibrate.
            RET

;------------------------------------------------------------------------------

            ;explanation here

STOP_CCD    LDP   #0
            SPLK  #10h,TCR          ;stop timer immediately.
            RET

*******************************************************************************
*                             MAIN PROGRAM                                    *
*******************************************************************************

            ;explanation of main program

MAIN        CALL INITIALISE
            ;CALL START_CCD

HERE        CALL  BEEPS
            IDLE                  ;wait for interrupts.

            .entry MAIN
            .end

;------------------------------------------------------------------------------

