/**
* Mk.c - make the flick translation table.
*
* Program to check and convert file specified on command line
* to one suitable for Flick rule conversion.
*
* What it does is:
*
*  - Strip out all of the comments (ie. lines starting with colon).
*
*  - Write out the version number and header code. (4 bytes)
*
*  - Compress the data for the 256 characters into a suitable form.
     This data contains:
                the input mapping to be done (eg. ASCII 187->space).
                the mapping for lower to upper case letters for this character set
                8 bits set for the user's choice (see fl_btdat.txt for example)
                Note: bits 3 and 6 are used for white space and wild cards
                      respectively.  Other bits can be used as required.
*
*  - The number of wildcards used. (1 byte).
*
*  - The data for the wildcards.
*       Format for Wildcards is:
*               wildcard character
*               whether to match one character, one or more characters,
*                                or zero or more characters
*               the bitmap to match against (as set for character data above)
*       3 bytes times number of wildcards.
*
*  - The number of input classes and states for decision table. (2 bytes)
*
*  - The decision table.  (number of input classes times number of states)
*
*  - The rules for convertion with the last line being just #
*    The tabs are removed from each rule and they are output in the format:
*       input class.
*       length of the rule (to speed up searching).
*       the body of the rule in the format:
*                {left context}[focus]{right context}=output.
*       a nul at the end of each rule.
*       the new state.
*
*  Version numbers
*
*      0.0  Still working on it.
*      0.1  Use binary file for output.
*
*  by Paul Blenkhorn
*
**/
 
#define         MATCH_SEVERAL           3
#define         MATCH_ONE               2
#define         MATCH_NONE              1
 
#define         BUFF_SIZE             256
 
#define         TAB                     9
 
/* standard IO library */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
 
void get_line(unsigned char buffer[]);
void do_version_number(void);
void get_char_data(void);
void get_wildcards(void);
void get_decision(void);
void get_main_table(void);
 
FILE *fp, *fo;                  /* bit dirty - but lot easier to have these global */
 
int main(argc,argv)
char *argv[];
{
    printf("MK: Flick Table Converter V0.1\n\n");
    printf("really\n");
 
    /* get input file name from the command line */
    if((fp=fopen(argv[1],"r")) == NULL) {
        printf("Input file %s not found",argv[1]);
        return(1);
    }
 
    /* get output file name from the command line */
    if((fo=fopen(argv[2],"wb")) == NULL) {
        printf("Output file %s not found",argv[2]);
        return(2);
    }
 
    do_version_number();
    get_char_data();
    get_wildcards();
    get_decision();
    get_main_table();
}
 
void do_version_number(void)
{
    unsigned char line[BUFF_SIZE];
 
    get_line(line); /* read version number & identifier */
    puts(line);
    fputc(17,fo);      /* upper case version of character       */
    fputc(12,fo);
    fputc(8,fo);
    fputc(atoi(line),fo);
}
 
void get_char_data(void)
{
    char line[BUFF_SIZE];
    int i;
    unsigned char flags;
 
    printf("Reading character data.\n");
    for (i = 0; i <= 255; i++) { /* read data for 256 characters */
        get_line(line);
 
        fputc(line[4],fo);      /* input transformation */
        fputc(line[5],fo);      /* upper case version of output character */
 
        flags = 0;              /* now work out how to set the flags    */
 
        if (  (line[7]  != '_' && !isalpha(line[7])) ||
              (line[8]  != '_' && !isalpha(line[7])) ||
              (line[9]  != '_' && !isalpha(line[8])) ||
              (line[10] != '_' && !isalpha(line[9])) ||
              (line[11] != '_' && !isalpha(line[10])) ||
              (line[12] != '_' && !isalpha(line[11])) ||
              (line[13] != '_' && !isalpha(line[12])) ||
              (line[14] != '_' && !isalpha(line[13])))
                printf("Error in character data %s.\n",line);
 
        if (isupper(line[7]))
              flags |= 1;
 
        if (isupper(line[8]))
              flags |= 2;
 
        if (isupper(line[9]))
              flags |= 4;
 
        if (isupper(line[10]))
              flags |= 8;
 
        if (isupper(line[11]))
              flags |= 16;
 
        if (isupper(line[12]))
              flags |= 32;
 
        if (isupper(line[13]))
              flags |= 64;
 
        if (isupper(line[14]))
              flags |= 128;
 
        fputc(flags,fo);
    }
}
 
void get_wildcards(void)
{
    char line[BUFF_SIZE];
    int wilds;
    int i;
 
    /* now read wildcards */
    get_line(line);
    wilds = atoi(line);
    fputc(wilds,fo);
 
    printf("Reading %d wildcard(s).\n",wilds);
 
    if (wilds > 0) { /* do wildcards */
        for (i = 0; i < wilds; i++) {
                get_line(line);
                fputc(line[1],fo);
                switch (line[3]) {
                        case '1' : /* match at least one character */
                                if (line[4] == '+')
                                        fputc(MATCH_SEVERAL,fo);
                                else
                                        fputc(MATCH_ONE,fo);
                                break;
                        case '0' : /* match at least no characters */
                                fputc(MATCH_NONE,fo);
                                break;
                        default : /* should not happen */
                                printf("Invalid wildcard specification: %s\n",line);
                                break;
                } /* switch */
        fputc(atoi(&line[6]),fo);  /* output the bit pattern for the wildcard */
        } /* for i = 0 to wilds */
    }
}
 
void get_decision(void)
{
    char line[BUFF_SIZE];
    int states;
    int input_classes;
    int i, j;
 
    /* now read no. of states */
    get_line(line);
    states = atoi(line);
    if (states <= 0)
        printf("Error in specifying no. of states: %s",line);
    fputc(states,fo);
 
    /* now read no. of input classes */
    get_line(line);
    input_classes = atoi(line);
    if (input_classes <= 0)
        printf("Error in specifying no. of input classes: %s",line);
    fputc(input_classes,fo);
 
    printf("Reading decision table %d by %d.\n",states, input_classes);
    for (i = 0; i < states; i++) {
        get_line(line);
        for (j = 0; j < input_classes; j++)
                if (line[j] == '0')
                        fputc(0,fo);
                else
                        fputc(1,fo);
    }
 
}
 
void get_main_table()
{
    char line[BUFF_SIZE];
    int i;
    int number,
        length,
        start,
        end;
 
    printf("Reading main rules.\n");
 
    get_line(line);
    while (line[0] != '#') {
        length = strlen(line);
 
        /* get the input class from the start of the line               */
        if (!isdigit(line[0]))
                printf("Invalid input class: %s\n",line);
 
        number = line[0] - '0';
        if (isdigit(line[1]))
                number = (number*10) + line[1] - '0';
 
        /* output the input class                                       */
        fputc(number,fo);
 
        /* get the length of the text to match - between the tabs       */
        start = 1;  /* find the start of it */
        while (line[start] == TAB) {
                start++;
                if (start > length)
                        printf("Invalid line: %s",line);
        }
 
        end = start; /* find end of string */
        while (line[end] != TAB) {
                end++;
                if (end > length)
                        printf("Invalid line: %s",line);
        }
        /* output the length of the text to match + NUL + CR        */
        putc(end-start+4,fo);   /* the length of the line */
        /* now output the text to match */
        for (i=start; i<end; i++)
                  fputc(line[i],fo);
        fputc(0,fo);
 
        /* get the new state from the end of the line                   */
        /* output the new state                                         */
        while (line[end] == TAB) {
                end++;
                if (end > length)
                        printf("Invalid line: %s",line);
        }
        if (isdigit(line[end]))
                number = line[end] - '0';
        else
                number = '-';
        if (isdigit(line[end+1]))
                number = (number*10) + line[end] - '0';
 
        /* output the new state                                       */
        fputc(number,fo);
 
        get_line(line);
    } /* while (line[0] != '#') */
 
    putc('#',fo);
    printf("Done.\n");
}
 
 
/* read the line in */
void get_line(unsigned char buffer[])
{
    int i = 0;
    int chr;
 
    do {  /* read in line not starting with colon (':') */
        i = 0;
        while ((chr = getc(fp)) != '\n') {
                 if (chr == EOF) {
                        printf("Premature end of file.\n");
                        exit(1);
                 }
                 buffer[i++] = chr;
        }
        buffer[i] = 0;
    } while (buffer[0] == ':');
}
 