/*
* Title:    Conv.c
*
* Purpose:  to decode a file (eg. braille to text, text to braille,
*           microwriter to text)
*
* Paul Blenkhorn
*
* Version:
*     0 - still working at it.
*     1 - 17/8/92.  First released version.
*     1(a) - 27/8/92.  Extract main functions and take them outside.
*                      Take out output buffer routine.
*     1(a)i - 31-8-92. Increment up_to for MATCH_ONE.
*     1(b)    27-4-93. Make table binary file.
*
*
* Notes:
*    1. Main tables are loaded by: tables().
*    2. Initialisation of arrays etc for each translation is performed by:
*         main_initialise()
*    3. Text is presented to convert() which return how many characters
*       it has translated.
*       For grade II braille to text this will be a word.
*    For text to grade II braille this will noramlly be a word
*       (sometimes more eg. AND WITH).
*       For Microwriting this will be a character at a time.
*    4. To make the system multi-lingual the information for the character
*       set, the code, wildcards, decision tables and exceptions table are all
*       configurable.  These are read in from the file passed to
*       main_initialise(), having been constructed by MK.EXE.
*       More information on this can be found in FL_BTDAT.TXT and MK.C.
*
*    Structure of file FLTAB
*    This data contains:
*       The input mapping to be done (eg. ASCII 187->space).
*       The mapping for lower to upper case letters for this character set
*       8 bits set for the user's choice (see fl_btdat.txt for example)
*          Note: bits 3 and 6 are used for white space and wild cards
*                respectively.  Other bits can be used as required.
*
*       The number of wildcards used. (1 byte).
*       The data for the wildcards.
*          Format for Wildcards is:
*               wildcard character
*               whether to match one character, one or more characters,
*                                or zero or more characters
*               the bitmap to match against (as set for character data above)
*               ie. 3 bytes times number of wildcards.
*
*       The number of input classes and states for decision table. (2 bytes)
*
*       The decision table.  (number of input classes times number of states)
*
*       The rules for convertion with the last line being just #
*       The tabs are removed from each rule and they are output in the format:
*               input class.
*               length of the rule (to speed up searching).
*               the body of the rule in the format:
*                       {left context}[focus]{right context}=output.
*               a NUL at the end of each rule.
*               the new state.
*
*
* General algorithm for convert:
*
*       initialise and load tables
*       while not end of input do
*       begin
*         input text is read in
*         while still converting do
*           begin
*           start at rule defined by current character
*           match = FALSE
*           do
*             if focus matches
*                  and state is ok
*                  and right context is ok
*                  and left context is ok
*               begin
*               output right hand side of rule
*               set new state
*               match = TRUE
*               move along input buffer
*               end
*             else go to next rule
*           until match
*           end
*       end
*
**/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
 
#define         STR_SIZE               256      /* number of characters in strings */
#define         TAB_SIZE             25000      /* size of tables */
 
#define         NUL                    '\0'

#define         FALSE                    0      /* logical constants */
#define         TRUE                     !FALSE

#define         NO_MOVE                  0
#define         NOT_DEFINED              0
 
#define         NUMBER_OF_CHARACTERS   256
 
#define         NO_FILE                 15
#define         VERSION_FAULT           16      /* error conditions */
#define         CHAR_FAULT              17
#define         WILD_FAULT              18
#define         DECISION_FAULT          19
#define         SIZE_FAULT              20
 
#define         MATCH_SEVERAL            3
#define         MATCH_ONE                2
#define         MATCH_NONE               1
 
#define         WILD_MATCH               1
#define         WILD_BIT_PATTERN         2
#define         WILD_DATA                3
 
#define         SPACE_BIT                8
#define         WILD_BIT                64
 
char
   caps,
   caps_lock;                       /* caps word flag */
 
unsigned char
   table[TAB_SIZE],                 /* area to hold exceptions table */
   current_state = 1;               /* the current state */
 
unsigned char
    *wild_tab,                      /* wildcard table */
    *decision_table;                /* state table */
 
unsigned int
    no_wilds,                       /* number of wildcards */
    no_input_classes,               /* number of input classes */
    no_states;                      /* number of states */
 
 
struct {        /* info for character set */
        unsigned char input_trans;      /* mapping for input chars */
        unsigned char to_up;            /* lower to upper case conversion */
        unsigned char data;             /* data for if chars are punc etc. */
        unsigned int hash;              /* position for hashing into tables */
} ch_info[NUMBER_OF_CHARACTERS];
 
/**************************** function prototypes ***************************/
static int find_match(int up_to, unsigned char input_dat[]);
static int still_converting(int convert_to, int count, unsigned char input_dat[]);
static int words_match(int up_to, unsigned char input_txt[]);
static int check_state(unsigned char inp_class);
static int right_context(int up_to, unsigned char input_dat[]);
static int left_context(int up_to, unsigned char input_dat[]);
static int wild_match(int step, int up_to, unsigned char input_dat[]);
static void initialise(void);
static void next_entry(unsigned char *this_table_entry);
static void match_found(void);
static void read_version_number(FILE *fp);
static void read_main_tables(FILE *fp);
static void read_character_data(FILE *fp);
static void read_decision(FILE *fp);
static void read_wildcards(FILE *fp);
void add_to_output(unsigned char chr);  /* in user programs - not here */
/************************ end function prototypes ***************************/
 
 
/* Function: convert
   Translate the text in the buffer input_dat.
   Parameters:
        convert_to is set for how many chars, or character type to convert to
        input_dat is the input data
   Returns:
        number of characters converted
*/
int convert(int convert_to, unsigned char input_dat[])
{
  int up_to = 0;                /* position in input buffer */
  int step;                     /* amount to step along input buffer */
#ifdef TESTING
printf("convert: ");
#endif
  initialise();
 
  do {
     /* check the table, return how far to move along input buffer
     if no match then move 1 char along the input buffer. */
 
     if (step = find_match(up_to,input_dat))
                up_to += step;
     else { /* output input character and change state to 1 */
        add_to_output(input_dat[up_to]);
        current_state = 1;
        up_to++;
     }
  } while (still_converting(convert_to,up_to,input_dat));
#ifdef TESTING
printf("\nconverted up to %d.",up_to);
#endif
  return(up_to);
}
 
 
/* Function: still_converting
   Have enough characters been converted - reached end condition?
   Parameters:
        convert_to is set for:
                how many chars (if < 32)
                specific character to convert to
             of convert to NUL (if convert_to == 0)
        up_to is current position in input data
        input_dat is the input text
   Returns:
        whether end condition satisfied
*/
static int still_converting(int convert_to,int up_to,unsigned char input_dat[])
{
#ifdef TESTING
printf("still converting?: up_to= %d convert to <%c>.\n",up_to,convert_to);
printf("ch <%c> info = <%d>.",input_dat[up_to], ch_info[input_dat[up_to]].data);
#endif
  if (input_dat[up_to] == 0)
        return(FALSE);
  if (convert_to == 0)               /* do till end of buffer */
        return(TRUE);
  if (convert_to < 32)               /* convert fixed number of characters */
        return(up_to > convert_to);
  else if (convert_to == ' ') {      /* convert to white space character */
#ifdef TESTING
printf(" space ");
#endif
        return(!(ch_info[input_dat[up_to]].data & SPACE_BIT));
  }
  else                               /* convert to specific character */
        return(convert_to == input_dat[up_to]);
}
 
 
/* Function: initialise
   Initialise for each group of characters to convert.
   Parameters:
        none
   Returns:
        void
*/
static void initialise(void)
{
  caps = caps_lock = FALSE;
}
 
/********************************* find match ******************************/
unsigned char
    *looking;                       /* pointer for search through tables */
 
/* Function: find_match
   Try to find a match in table for current position in input buffer from
   tables.
   Parameters:
        up_to is position in input buffer
        input_dat is the input data
   Returns:
        number of characters converted - NO_MOVE (0) if none
*/
static int find_match(int up_to, unsigned char input_dat[])
  {
  int move_no;                          /* how far to move along input buffer */
  unsigned char *this_table_entry;      /* pointer to input class of current table entry */
 
#ifdef TESTING
printf("find match: ");
#endif
 
  /* quick hash into the contraction table from first character in buffer.
      *looking == 0 if no entry found.
  */
  this_table_entry = looking = &table[ch_info[input_dat[up_to]].hash];
 
#ifdef TESTING
printf("character %c table pos = %d ",input_dat[up_to],ch_info[input_dat[up_to]].hash);
#endif
 
    /* if hash character then check rules */
  if (*looking == NOT_DEFINED)
        return(NO_MOVE);
  else {
        do {  /* go through the table entries */
           while (*looking++ != '[')
                   ; /* get to character after '[' */
 
#ifdef TESTING
printf("<%s>",looking);
#endif
           if (*looking != input_dat[up_to]) /* run out of table entries for this letter? */
                        return(NO_MOVE);
 
           if (move_no = words_match(up_to,input_dat))
                if (check_state(*(this_table_entry-1)))
                        if (right_context(up_to+move_no,input_dat))
                                if (left_context(up_to,input_dat)) {
                                        match_found();
                                        return(move_no);
                                }
 
    /* go to next entry in the table */
           looking = (this_table_entry += *this_table_entry);
 
        } while(TRUE);
  } /* *looking */
}
 
 
/* Function: check_state
   check the input class against the current state in the state table
   Parameters:
        chr is the input class
   Returns:
        whether input class is acceptable for match
*/
static int check_state(unsigned char inp_class)
{
#ifdef TESTING
printf("in. cl.=%d state = %d.",inp_class,current_state);
#endif
  return(decision_table[(no_input_classes*(current_state-1)) + inp_class-1]);
}
 
 
/* Function: left_context
   Check to see if the left context of the match string is valid.
   Parameters:
        up_to is end of match string in input data
        input_dat is the input text
   Returns:
        whether left context is satisfied
*/
static int left_context(int up_to,unsigned char input_dat[])
{
#ifdef TESTING
printf("check left: ");
#endif
    while (*looking-- != '[')
       ;
    return(wild_match(-1,--up_to,input_dat));
}
 
 
/* Function: right_context
   Check to see if the right context of the match string is valid.
   Parameters:
        up_to is end of match string in input data
        input_dat is the input text
   Returns:
        whether right context is satisfied
*/
static int right_context(int up_to,unsigned char input_dat[])
{
#ifdef TESTING
printf("check right: ");
#endif
    looking++;
    return(wild_match(+1,up_to,input_dat));
}
 
 
/* Function: wild_match
   Used by left_context and right_context to check contexts.
   It uses wild cards and other characters to validate context
   Parameters:
        step is +ve or -ve to tell function which way to increment
        up_to is end of match string in input data
        input_dat is the input text
   Returns:
        whether match is successful
*/
static int wild_match(int step,int up_to,unsigned char input_dat[])
{
  register int i;
  unsigned char bits;
 
#ifdef TESTING
printf("wild match ");
#endif
  while (*looking >= ' ') { /* work through the entries */
     if (*looking == '=' && step == 1)
        break;
     if (ch_info[*looking].data & WILD_BIT) { /* got wildcard */
        for (i = 0; i < no_wilds; i++)
           if (*looking == wild_tab[i*WILD_DATA]) { /* found wild card */
              bits = wild_tab[i*WILD_DATA+WILD_BIT_PATTERN];
              switch (wild_tab[i*WILD_DATA+WILD_MATCH]) {
                case MATCH_ONE :
                  if (!(ch_info[input_dat[up_to]].data & bits))
                        return(FALSE);
                  up_to += step;
                  break;
                case MATCH_SEVERAL :
                  if (!(ch_info[input_dat[up_to]].data & bits))
                        return(FALSE);
                  do {
                        up_to += step;
                  } while (ch_info[input_dat[up_to]].data & bits);
                  break;
                case MATCH_NONE :
                  while (ch_info[input_dat[up_to]].data & bits)
                        up_to += step;
                  break;
              } /* switch */
           break;
           } /* (if) found wild card */
     } /* got wildcard */
     else { /* not wildcard */
        if (*looking != input_dat[up_to])
           return(FALSE);
        up_to += step;
     }   /* not wildcard */
     looking += step;
  } /* while work through the entries */
  return(TRUE);
}
 
 
/* Function: words_match
   compare input buffer with focus of entry in table [in brackets]
   Parameters:
        up_to is end of match string in input data
        input_txt is the input text
   Returns:
        whether words do match
*/
static int words_match(int up_to,unsigned char input_txt[])
{
  int start = up_to;
 
#ifdef TESTING
printf("words match: ");
#endif
  do {
     if (*looking++ != input_txt[up_to++])
           return(FALSE);
  } while (*looking != ']');
#ifdef TESTING
printf("yes:");
#endif
  return(up_to-start);
}
 
 
/* Function: match_found
   move the transcribed text (right hand side of the rule) to the output buffer.
   update the state of the system.
   Parameters:
        none
   Returns:
        void
*/
static void match_found(void)
{
#ifdef TESTING
printf("match found: ");
#endif
  while (*looking++ != ']')     /* get to rhs of rule */
        ;
  while (*looking++ != '=')     /* get to rhs of rule */
        ;
  while (*looking)              /* output info */
     add_to_output(*looking++);
 
  ++looking;
  if (*looking != '-')  /* ie. no state change */
        current_state = *looking;  /* update state */
#ifdef TESTING
printf("New state %d:",current_state);
#endif
}
/****************************** end find match ******************************/
 
/********************************* read tables ******************************/
 
 
/* Function: main_initialise
   main initialisation for program - read in tables from disc.
   Parameters:
        name    the name of the file.
   Returns:
        nothing
*/
void main_initialise(char *name)
  {
  FILE *fp;
 
  if ( (fp = fopen(name,"rb")) == NULL) {
      printf("Fatal error! Cannot find file %s!",name);
      exit(NO_FILE);
  }
 
  read_version_number(fp);
  read_character_data(fp);
  read_wildcards(fp);
  read_decision(fp);
  read_main_tables(fp);
 
  fclose(fp);
}
 
 
/* Function: table_fault
   fault occured in reading table.  Report fault and exit.
   Parameters:
        fault_no the fault to be reported
   Returns:
        whether left context is satisfied
*/
static void table_fault(int fault_no)
{
    printf("\n");
        switch (fault_no) {
          case VERSION_FAULT:
        printf("%%Error in reading version number!\n");
                break;
          case CHAR_FAULT:
        printf("%%Error in reading character data!\n");
                break;
          case WILD_FAULT:
        printf("%%Error in reading wild card data!\n");
                break;
          case DECISION_FAULT:
        printf("%%Error in reading decision tables!\n");
                break;
          case SIZE_FAULT:
        printf("%%Error in reading rule tables!\n");
                break;
        }
    printf("Aborting.\n");
        exit(fault_no);
}
 
 
/* Function: read_version_number
   read and verify version number from program data file
   Parameters:
        fp handle for file
   Returns:
        void
*/
static void read_version_number(FILE *fp)
{
        int version_number;
 
        if (fgetc(fp) != 17)
                table_fault(VERSION_FAULT);
        if (fgetc(fp) != 12)
                table_fault(VERSION_FAULT);
        if (fgetc(fp) != 8)
                table_fault(VERSION_FAULT);
        version_number = fgetc(fp);
#ifdef TESTING
printf("%d:",version_number);
#endif
}
 
 
/* Function: read_character_data
   read and verify data for 256 characters.
   Parameters:
        fp handle for file
   Returns:
        void
*/
static void read_character_data(FILE *fp)
{
  register int i;
  register int chr;
  for (i = 0; i < 256; i++) {
        if ((chr = fgetc(fp)) == EOF)
                table_fault(CHAR_FAULT);
        else
            ch_info[i].input_trans = chr;
        if ((chr = fgetc(fp)) == EOF)
                table_fault(CHAR_FAULT);
        else
            ch_info[i].to_up = chr;
        if ((chr = fgetc(fp)) == EOF)
                table_fault(CHAR_FAULT);
        else
            ch_info[i].data = chr;
  }
 
}
 
 
/* Function: read_wildcards
   read and verify number of wildcards from program data file.
   Allocate memory and read and verify wildcard data.
   Parameters:
        fp handle for file
   Returns:
        void
*/
static void read_wildcards(FILE *fp)
{
   register int i;
   register int chr;
 
   if ( (no_wilds = fgetc(fp)) == EOF)
        table_fault(WILD_FAULT);
   wild_tab = (char *) malloc(no_wilds*WILD_DATA);
#ifdef TESTING
printf("Wildcards:");
#endif
   for (i = 0; i < no_wilds*WILD_DATA; i++) {
        if ( (chr = fgetc(fp)) == EOF)
                table_fault(WILD_FAULT);
        wild_tab[i] = chr;
#ifdef TESTING
printf("%c",chr);
#endif
   }
#ifdef TESTING
printf("\n");
#endif
}
 
 
/* Function: read_decision table
   read and verify number of states and input classes from program data file.
   Allocate memory and read and verify decision table data.
   Parameters:
        fp handle for file
   Returns:
        void
*/
static void read_decision(FILE *fp)
{
   int i, j, chr;
 
   if ( (no_states = fgetc(fp)) == EOF)
        table_fault(DECISION_FAULT);
   if ( (no_input_classes = fgetc(fp)) == EOF)
        table_fault(DECISION_FAULT);
   decision_table = (char *) malloc(no_input_classes*no_states);
   for (i = 0; i < no_states; i++) {
#ifdef TESTING
printf("\n%d: ",i);
#endif
        for (j = 0; j < no_input_classes; j++) {
                if ( (chr = fgetc(fp)) == EOF)
                        table_fault(DECISION_FAULT);
                decision_table[i*no_input_classes+j] = chr;
#ifdef TESTING
printf("%d",chr);
#endif
        }
   }
}
 
 
/* Function: read_main tables
   read into table[] and verify exceptions from program data file.
   Build hash into table[] based on first character of focus [in brackets].
   Parameters:
        fp handle for file
   Returns:
        void
*/
static void read_main_tables(FILE *fp)
{
  register int i;
  register int chr;
  int start;
 
  for (i = 0; i < 256; i++)             /* initialise hash table */
        ch_info[i].hash = NOT_DEFINED;
 
  i = 0;
  table[i++] = 0;       /* at start of table have dummy NUL and new state */
  table[i++] = 1;       /* to ensure that algorithms work */
 
  while (!feof(fp)) {
        table[i++] = fgetc(fp);                         /* input class */
        start = i;
        table[i++] = fgetc(fp);                   /* length of entry */
        if (feof(fp))
                break;
        while ( (table[i++] = fgetc(fp)) != '[' ) /* skip over left context and '[' */
                ;
        table[i++] = chr = fgetc(fp);             /* first character of focus */
        if (ch_info[chr].hash == NOT_DEFINED) {
#ifdef TESTING
printf("%c%d:",chr,start);
#endif
              ch_info[chr].hash = start;
        }
 
        /* read to the end of the line */
        while (table[i++] = fgetc(fp))
                if (feof(fp))
                        break;
        table[i++] = fgetc(fp);                   /* new state */
        if (i >= TAB_SIZE)
                table_fault(SIZE_FAULT);
  }
 
}
 
/***************************** end read tables ******************************/
 
 