/**
* Mn.c - make the flick normalising table.
*
* Program to check and convert file specified on command line
* to one suitable for Flick normalising.
*
* What it does is:
*
*  - Strip out all of the comments (ie. lines starting with colon).
*
*  - Write out the version number and header code. (4 bytes)
*
*  - Read number of inputs used. (1 byte).
*       Then the labels used to define those inputs.
*
*  - The number of sound inputs used for feedback
*       Then the values of those sound inputs
*
*  - The rules for normalisation with the last line being just #
*    The tabs are removed from each rule and they are output in the format:
*       Sound feedback used (0 for none).
*       length of the rule (to speed up searching).
*       the body of the rule in the format:
*                byte,byte=output
*       (each byte is built out of the bits set by the switches)
*       a nul at the end of each rule.
*
*  Version numbers
*
*      0.0  Still working on it.
*
*  by Paul Blenkhorn
*
**/
 
#define         BUFF_SIZE             256
 
#define         TAB                     9
 
/* standard IO library */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
 
int
        no_of_inputs,           /* how many inputs are used */
        no_of_sounds;           /* how many sounds are used */
 
char    input_data[BUFF_SIZE];  /* buffer for switch labels */
 
void get_line(unsigned char buffer[]);
void do_version_number(void);
void get_mode(void);
void get_input_data(void);
void get_sound_data(void);
void get_main_table(void);
 
FILE *fp, *fo;                  /* bit dirty - but lot easier to have these global */
 
int main(argc,argv)
char *argv[];
{
    printf("MK: Flick Table Converter V0.0\n\n");
 
    /* get input file name from the command line */
    if((fp=fopen(argv[1],"r")) == NULL) {
        printf("Input file %s not found",argv[1]);
        return(1);
    }
 
    /* get output file name from the command line */
    if((fo=fopen(argv[2],"w")) == NULL) {
        printf("Output file %s not found",argv[2]);
        return(2);
    }
 
    do_version_number();
    get_mode();
    get_input_data();
    get_sound_data();
    get_main_table();
}
 
 
void do_version_number(void)
{
    unsigned char line[BUFF_SIZE];
    int number;
 
    get_line(line);             /* read version number & identifier */
    printf("Version: ");
    puts(line);
    fputc(18,fo);               /* upper case version of character       */
    fputc(13,fo);
    fputc(9,fo);
    fputc(atoi(line),fo);
}
 
void get_mode(void)
{
    unsigned char line[BUFF_SIZE];
    int number;
 
    get_line(line);             /* read version number & identifier */
    fputc(line[0]-'0',fo);               /* upper case version of character       */
}
 
 
void get_input_data(void)
{
    char line[BUFF_SIZE];
 
    get_line(line);                     /* get number of inputs */
 
    no_of_inputs = atoi(line);
    fputc(no_of_inputs,fo);
    printf("Reading %d inputs(s).\n", no_of_inputs);
    if (no_of_inputs == 0) {
        printf("No inputs specified -- aborting.\n");
        exit(16);
    }
 
    get_line(input_data);               /* read labels used for inputs */
}
 
 
void get_sound_data(void)
{
    char line[BUFF_SIZE];
    int sound;
    int i;
 
    get_line(line);                     /* get number of sounds */
 
    no_of_sounds = atoi(line);
    fputc(no_of_sounds,fo);
    printf("Reading %d sound(s).\n", no_of_sounds);
    if (no_of_sounds == 0)
        return;
 
    for (i = 0; i < no_of_sounds; i++) {
        get_line(line);                         /* read each number for sounds */
        sound = atoi(line);
        putw(sound,fo);
    }
}
 
 
/** Name        get_main_table
*               read each line of data from disc and output it
*               after processing
*
*   Parameters:
*       none
*
*   Returns:
*       nothing
**/
void get_main_table()
{
    unsigned char
        line[BUFF_SIZE],                /* line read from file */
        out_line[BUFF_SIZE];            /* output buffer to build line */
 
    int
        i,
        bits,                           /* bits set for current group */
        flag,                           /* bit set for current switch label */
        up_to,                          /* where we are up in in 'line' */
        s_number,                       /* the sound number */
        length,                         /* length of input buffer */
        out_length;                     /* position so far in output buffer */
 
    printf("Reading main rules.\n");
 
    do { /* while (line[0] != '#') */
        get_line(line);
        out_length = 0;
        if (line[0] == '#')
                break;
        length = strlen(line);
 
        /* get the sound number from the start of the line               */
        if (!isdigit(line[0]))
            fputc(0,fo);
        else {
            s_number = line[0] - '0';
            if (isdigit(line[1]))
                s_number = (s_number*10) + line[1] - '0';
 
            fputc(s_number,fo);                                 /* output the sound number  */
        }
 
        up_to = 1;  /* find the start of input settings */
        while (line[up_to] != '[') {
            up_to++;
            if (up_to >= length)
                printf("Invalid line: %s",line);
        }
 
        /* work through groups separated by commas until get to ']' */
        while (line[up_to] != ']') {
            up_to++;
            bits = 0;
            while (line[up_to] != ',' && line[up_to] != ']') {  /* get group */
                for (flag = 0, i = 0; i < no_of_inputs; i++) {  /* find each match */
                    if (input_data[i] == line[up_to]) {
                        flag = (1 << i);                /* set bit */
                        break;
                    }
                } /* for i < no_of_inputs */
                if (flag == 0)
                    printf("Error <%c> in switch label %s.\n",line[up_to],line);
                bits |= flag;                           /* or in bits for input */
                if (up_to >= length) {                  /* error check */
                    printf("No ']' for rule: %s.\n",line);
                    break;
                }
                up_to++;
            } /* while != ',' or ']' */
            out_line[out_length++] = bits;
 
            if (up_to > length) {                       /* error check */
                break;
            }
            if (line[up_to] == ',')
                out_line[out_length++] = ',';           /* store seperator */
            else
                out_line[out_length++] = '=';           /* ditto */
        } /* while != ']' */
        if (up_to++ >= length)
            continue;
 
        while (line[up_to] != '=') {                    /* now get '=' */
            up_to++;
            if (up_to >= length)                        /* error check */
                printf("No ']' for rule: %s.\n",line);
        }
        if (line[++up_to] != TAB)                       /* error check */
                printf("Tab not found after equals:%s\n",line);
 
        for (i = ++up_to; i < length-1; i++)            /* store rest of line */
            out_line[out_length++] = line[i];
        out_line[out_length] = 0;
        if (line[length-1] != '<')                      /* error check */
                printf("'<' not found at end of line:%s\n",line);
 
        fputc(out_length,fo);                            /* output line length */
        for (i = 0; i < out_length; i++)                 /* output line */
           fputc(out_line[i],fo);
        fputc(0,fo);
    } while (1); /* break if line[0] == '#' */
    fputc('#',fo);
    printf("Done.\n");
}
 
 
/** Name        get_line
*               read a line from disc
*               ignore lines which begin with colon
*   Parameters:
*       pointer to buffer to store line in
*
*   Returns:
*       nothing
**/
void get_line(unsigned char buffer[])
{
    int i = 0;
    int chr;
 
    do {  /* read in line not starting with colon (':') */
        i = 0;
        while ((chr = fgetc(fp)) != '\n') {
                 if (chr == EOF) {
                        printf("Premature end of file.\n");
                        printf("Last line should have '#'.\n");
                        exit(3);
                 }
                 buffer[i++] = chr;
        }
        buffer[i] = 0;
    } while (buffer[0] == ':');
}
 